/*
 * Copter Commander - A 2d network helicopter game
 * Copyright (C) 2001  David Morse
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _BSD_SOURCE /* hack to get strncasecmp under glibc */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alloca.h> /* for alloca. (not stdlib because of --ansi) */
#include <assert.h>
#include "coco.h"

unsigned short coco_parse_port( char *strng, char *cl_arg )
{
  long result;
  char *end;
  result = strtol(strng, &end, 0);
  if (!((result >= 0) && 
        (result < (1<<16)) && 
        ((end - strng) == strlen( strng ))))
    coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                        "Couldn't parse port \"%s\".\n", cl_arg );
  return result;
}


/* return the new index, i.e. idx+args_consumed */
int coco_process_shared_arguments( int idx, int argc, char **argv )
{
  if ((0==strcmp( "--version", argv[idx] )))
    coco_version_message();
  else
    if ((0==strcmp( "-?", argv[idx] )) || (0==strcmp( "--help", argv[idx] )))
      coco_help_message();
  else
    if (0==strcmp( "--usage", argv[idx] ))
      coco_usage_message( COCO_TRUE );
  else
    if (0==strcmp( "--list-levels", argv[idx] ))
      coco_list_levels_message();
  else
    if ((0==strcmp( "-p", argv[idx] )) || (0==strcmp( "--port", argv[idx] )))
      {
        idx++;
        if (idx<argc)
          coco_port = coco_parse_port( argv[idx], argv[idx] );
        else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                 "Missing argument to \"%s\" option.\n",
                                 argv[idx-1] );
        return idx+1;
      }
  else if (0==strncmp( "--port=", argv[idx], 7 ))
    { 
      coco_port = coco_parse_port( argv[idx] + 7, argv[idx] );
      return idx+1;
    }
  return idx;
}

/* coco_mallocs return value */
char *
coco_sloppy_filename_to_symname( const char *levelname )
{
  int i;
  int len = strlen( levelname );
  int end = len;
  int start = 0;
  int cpylen;
  char *cpy;

  for (i=len-1; i>=0; i--)
    if (levelname[i] == '.')
      {
        end = i-1;
        break;
      }
  if (i<0) 
    i = len-1;
  else 
    i = i; /* i NOT reset */
  for (; i>=0; i--)
    if (levelname[i] == COCO_DIRECTORY_SEPARATOR_CHAR)
      {
        start = i+1;
        break;
      }
  cpylen = end - start + 1;
  cpy = coco_malloc( cpylen );
  memcpy( cpy, levelname + start, cpylen );
  return cpy;
}

unsigned long coco_command_line_left_comps = 0;
unsigned long coco_command_line_right_comps = 0;

/* Returns true if the main program should run */
coco_bool 
coco_process_server_arguments( int argc, char **argv )
{
  int j, i=0;
  coco_bool tutorial = COCO_FALSE;
  coco_bool level_file_name_malloced_p = COCO_FALSE;
  char *log_file_name = NULL;
  char *level_file_name = NULL;
  argc-=2; 
  argv+=2;
  while (i < argc)
    {
      j = coco_process_shared_arguments( i, argc, argv );
      if (j!=i)
        { i=j; continue; }
      else if ((0==strcmp( "--log", argv[i] )))
        {
          i++;
          if (i<argc)
            log_file_name = argv[i];
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                   "Missing argument to \"%s\" option.\n",
                                   argv[i-1] );
        }
      else if (0==strncmp( "--log=", argv[i], 10 ))
        log_file_name = argv[i] + 10;
      else if ((0==strcmp( "--left-comps", argv[i] )))
        {
          i++;
          if (i<argc)
            coco_command_line_left_comps = strtoul( argv[i], NULL, 10 );
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                   "Missing argument to \"%s\" option.\n",
                                   argv[i-1] );
        }
      else if (0==strncmp( "--left-comps=", argv[i], 13 ))
        {
          coco_command_line_left_comps = strtoul( argv[i] + 13, NULL, 10 );
          
        }
      else if ((0==strcmp( "--right-comps", argv[i] )))
        {
          i++;
          if (i<argc)
            coco_command_line_right_comps = strtoul( argv[i], NULL, 10 );
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                   "Missing argument to \"%s\" option.\n",
                                   argv[i-1] );
        }
      else if (0==strncmp( "--right-comps=", argv[i], 14 ))
        {
          coco_command_line_right_comps = strtoul( argv[i] + 14, NULL, 10 );
          
        }
      else if ((0==strcmp( "-q", argv[i] )) || 
               (0==strcmp( "--quiet", argv[i] )))
        log_file_name = "/dev/null";
      else if ((0==strcmp( "--level", argv[i] )))
        {
          i++;
          if (i<argc)
            level_file_name = argv[i];
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                   "Missing argument to \"%s\" option.\n",
                                   argv[i-1] );
        }
      else if (0==strncmp( "--level=", argv[i], 10 ))
        level_file_name = argv[i] + 10;
      else if (0==strcmp( "--tutorial", argv[i] ))
        tutorial = COCO_TRUE;
      else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                               "Unknown option \"%s\".\n",
                               argv[i] );
      i++;
    }
  coco_set_log_file( log_file_name );
  if (tutorial)
    {
      char *symname = coco_sloppy_filename_to_symname( level_file_name );
      int i;
      for (i=0; i<COCO_TUTORIAL_INFO_SIZE; i++)
        if (0==strcmp( symname, coco_tutorial_infos[i].name ))
          {
            coco_tutorial = coco_tutorial_infos + i;
            coco_free( symname );
            symname = NULL;
            break;
          }
        else coco_fatal_printf( COCO_ERROR_LEVEL_FILE_NOT_FOUND,
                                "\"%s\" can't be found in the "
                                "tutorial_info table.\n",
                                symname );
    }

  if (!coco_massage_level_file_name( &level_file_name,
                                     &level_file_name_malloced_p ))
    coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                       "Couldn't massage %s into a valid level file name.\n",
                       level_file_name );
  coco_level_text = coco_map_level_text( level_file_name, NULL );
  if (level_file_name_malloced_p) coco_free( level_file_name );
  return COCO_TRUE;
}

static coco_nationality
coco_parse_nationality( char *str )
{
  if (0==strcasecmp( str, COCO_BAD_GUYS_PRETTY_NAME ))
    return COCO_BAD_GUYS;
  else if (0==strcasecmp( str, COCO_GOOD_GUYS_PRETTY_NAME ))
    return COCO_GOOD_GUYS;
  else return COCO_NEUTRAL;
}

#define COCO_CHECK_SKY_OPTION( sky_opt ) \
          if ((sky_opt<0) || (sky_opt >= COCO_SKY_LIMIT)) \
            coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE, \
                               "--sky argument must be between 0 and %d, " \
                               "got %d\n", \
                               (int) COCO_SKY_LIMIT-1, sky_opt )



/* Returns true if the main program should run */
coco_bool 
coco_process_client_arguments( int argc, char **argv )
{
  int i, j;
  int gnome_arg_start = -1;
  int gnome_arg_end = -1;

  i=1;
  while (i<argc)
    {
      j = coco_process_shared_arguments( i, argc, argv );
      if (j!=i)
        { i=j; continue; }
      else if ((0==strcmp( "-s", argv[i] )) || 
               (0==strcmp( "--server", argv[i] )))
        {
          i++;
          if (i<argc)
            {
              coco_server = argv[i];
              coco_server_specified_p = COCO_TRUE;
            }
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                   "Missing argument to \"%s\" option.\n",
                                   argv[i-1] );
        }
      else if (0==strncmp( "--server=", argv[i], 9 ))
        {
          coco_server = argv[i] + 9;
          coco_server_specified_p = COCO_TRUE;
        }
      else if (0==strncmp( "--name=", argv[i], 7 ))
        {
          coco_set_nickname( argv[i] + 7 );
        }
      else if ((0==strcmp( "--name", argv[i] )))
        {
          i++;
          if (i<argc)
            coco_set_nickname( argv[i] );
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                  "Missing argument to \"%s\" option.\n",
                                  argv[i-1] );
        }
      else if (0==strncmp( "--team=", argv[i], 7 ))
        {
          coco_cl_nationality = coco_parse_nationality( argv[i] + 7 );
          if (coco_cl_nationality == COCO_NEUTRAL)
            coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                               "%s option must be either \"left\" or"
                               " \"right\".\n", "--team" );
        }
      else if ((0==strcmp( "--team", argv[i] )))
        {
          i++;
          if (i<argc)
            {
              coco_cl_nationality = coco_parse_nationality( argv[i] );
              if (coco_cl_nationality == COCO_NEUTRAL)
                coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                   "%s option must be either \"left\" or"
                                   " \"right\".\n", argv[i-1] );
            }
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                  "Missing argument to \"%s\" option.\n",
                                  argv[i-1] );
        }
      else if (0==strcmp( "--single-buffer", argv[i] ))
        coco_command_line_inhibit_double_buffer = COCO_TRUE;
      else if (0==strcmp( "--blend-alpha", argv[i] ))
        coco_blend_alpha = 1;
      else if (0==strcmp( "--no-blend-alpha", argv[i] ))
        coco_blend_alpha = 0;
      else if (0==strcmp( "--gnome-options", argv[i] ))
        {
          int k;
          for (k=i+1; (k<argc)&&strcmp( "--coco-options", argv[k] ); k++)
            ;
          gnome_arg_start = i+1;
          gnome_arg_end = k;
          i = k;
        }
      else if (0==strcmp( "--sky", argv[i] ))
        {
          i++;
          if (i<argc)
            coco_sky_option = strtoul( argv[i], NULL, 10 );
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                  "Missing argument to \"%s\" option.\n",
                                  argv[i-1] );
          COCO_CHECK_SKY_OPTION( coco_sky_option );
        }
      else if (0==strncmp( "--sky=", argv[i], 6 ))
        {
          coco_sky_option = strtoul( argv[i] + 6, NULL, 10 );
          COCO_CHECK_SKY_OPTION( coco_sky_option );
        }
      else if (0==strcmp( "--eye-candy", argv[i] ))
        {
          i++;
          if (i<argc)
            coco_eye_candy = strtoul( argv[i], NULL, 10 );
          else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                                  "Missing argument to \"%s\" option.\n",
                                  argv[i-1] );
          COCO_CHECK_SKY_OPTION( coco_sky_option );
        }
      else if (0==strncmp( "--eye-candy=", argv[i], 12 ))
        {
          coco_eye_candy = strtoul( argv[i] + 12, NULL, 10 );
        }
      else if (0==strcmp( "--no-gii", argv[i] ))
        coco_use_gii = COCO_FALSE;
      else coco_fatal_printf( COCO_ERROR_BAD_COMMAND_LINE,
                               "Unknown option \"%s\".\n",
                               argv[i] );
      i++;
    }
  coco_process_gnome_args( gnome_arg_start, gnome_arg_end, argv );
  return COCO_TRUE;
}

